@@ -9,14 +9,14 @@ class TestPatternMatching < Test::Unit::TestCase
   end
 
   def setup
-    if defined?(DidYouMean)
+    if defined?(DidYouMean.formatter=nil)
       @original_formatter = DidYouMean.formatter
       DidYouMean.formatter = NullFormatter.new
     end
   end
 
   def teardown
-    if defined?(DidYouMean)
+    if defined?(DidYouMean.formatter=nil)
       DidYouMean.formatter = @original_formatter
     end
   end
@@ -109,16 +109,12 @@ class TestPatternMatching < Test::Unit::TestCase
     end
 
     assert_block do
-      # suppress "warning: Pattern matching is experimental, and the behavior may change in future versions of Ruby!"
-      experimental, Warning[:experimental] = Warning[:experimental], false
       eval(%q{
         case true
         in a
           a
         end
       })
-    ensure
-      Warning[:experimental] = experimental
     end
 
     assert_block do
@@ -201,11 +197,49 @@ class TestPatternMatching < Test::Unit::TestCase
       end
     end
 
-    assert_syntax_error(%q{
+    assert_valid_syntax(%{
+      case 0
+      in [ :a | :b, x]
+        true
+      end
+    })
+
+    assert_in_out_err(['-c'], %q{
       case 0
       in a | 0
       end
-    }, /illegal variable in alternative pattern/)
+    }, [], /alternative pattern/,
+    success: false)
+
+    assert_in_out_err(['-c'], %q{
+      case 0
+      in 0 | a
+      end
+    }, [], /alternative pattern/,
+    success: false)
+  end
+
+  def test_alternative_pattern_nested
+    assert_in_out_err(['-c'], %q{
+      case 0
+      in [a] | 1
+      end
+    }, [], /alternative pattern/,
+    success: false)
+
+    assert_in_out_err(['-c'], %q{
+      case 0
+      in { a: b } | 1
+      end
+    }, [], /alternative pattern/,
+    success: false)
+
+    assert_in_out_err(['-c'], %q{
+      case 0
+      in [{ a: [{ b: [{ c: }] }] }] | 1
+      end
+    }, [], /alternative pattern/,
+    success: false)
   end
 
   def test_var_pattern
@@ -357,6 +391,14 @@ END
       end
     end
 
+    assert_block do
+      a = "abc"
+      case 'abc'
+      in /#{a}/o
+        true
+      end
+    end
+
     assert_block do
       case 0
       in ->(i) { i == 0 }
@@ -464,6 +506,8 @@ END
         true
       end
     end
+
+    assert_valid_syntax("1 in ^(1\n)")
   end
 
   def test_array_pattern
@@ -798,6 +842,10 @@ END
         true
       end
     end
+
+    assert_syntax_error(%q{
+      0 => [a, *a]
+    }, /duplicated variable name/)
   end
 
   def test_find_pattern
@@ -866,6 +914,10 @@ END
         false
       end
     end
+
+    assert_syntax_error(%q{
+      0 => [*a, a, b, *b]
+    }, /duplicated variable name/)
   end
 
   def test_hash_pattern
@@ -1155,7 +1207,7 @@ END
       end
     end
 
-    bug18890 = assert_warning(/(?:.*:[47]: warning: unused literal ignored\n){2}/) do
+    bug18890 = assert_warning(/(?:.*:[47]: warning: possibly useless use of a literal in void context\n){2}/) do
       eval("#{<<~';;;'}")
       proc do |i|
         case i
@@ -1317,7 +1369,7 @@ END
     end
 
     assert_block do
-      case {}
+      case C.new({})
       in {}
         C.keys == nil
       end
@@ -1570,6 +1622,18 @@ END
     assert_equal false, (1 in 2)
   end
 
+  def test_bug18990
+    {a: 0} => a:
+    assert_equal 0, a
+    {a: 0} => a:
+    assert_equal 0, a
+
+    {a: 0} in a:
+    assert_equal 0, a
+    {a: 0} in a:
+    assert_equal 0, a
+  end
+
   ################################################################
 
   def test_single_pattern_error_value_pattern
@@ -1638,7 +1702,7 @@ END
       raise a # suppress "unused variable: a" warning
     end
 
-    assert_raise_with_message(NoMatchingPatternKeyError, "{:a=>0}: key not found: :aa") do
+    assert_raise_with_message(NoMatchingPatternKeyError, "{a: 0}: key not found: :aa") do
       {a: 0} => {aa:}
       raise aa # suppress "unused variable: aa" warning
     rescue NoMatchingPatternKeyError => e
@@ -1647,7 +1711,7 @@ END
       raise e
     end
 
-    assert_raise_with_message(NoMatchingPatternKeyError, "{:a=>{:b=>0}}: key not found: :bb") do
+    assert_raise_with_message(NoMatchingPatternKeyError, "{a: {b: 0}}: key not found: :bb") do
       {a: {b: 0}} => {a: {bb:}}
       raise bb # suppress "unused variable: bb" warning
     rescue NoMatchingPatternKeyError => e
@@ -1656,15 +1720,15 @@ END
       raise e
     end
 
-    assert_raise_with_message(NoMatchingPatternError, "{:a=>0}: 1 === 0 does not return true") do
+    assert_raise_with_message(NoMatchingPatternError, "{a: 0}: 1 === 0 does not return true") do
       {a: 0} => {a: 1}
     end
 
-    assert_raise_with_message(NoMatchingPatternError, "{:a=>0}: {:a=>0} is not empty") do
+    assert_raise_with_message(NoMatchingPatternError, "{a: 0}: {a: 0} is not empty") do
       {a: 0} => {}
     end
 
-    assert_raise_with_message(NoMatchingPatternError, "[{:a=>0}]: rest of {:a=>0} is not empty") do
+    assert_raise_with_message(NoMatchingPatternError, "[{a: 0}]: rest of {a: 0} is not empty") do
       [{a: 0}] => [{**nil}]
     end
   end
