@@ -28,6 +28,7 @@ begin
 rescue LoadError
   $" << "zlib.rb"
 end
+require_relative 'lib/path'
 
 INDENT = " "*36
 STDOUT.sync = true
@@ -96,6 +97,20 @@ def parse_args(argv = ARGV)
   opt.on('--gnumake') {gnumake = true}
   opt.on('--debug-symbols=SUFFIX', /\w+/) {|name| $debug_symbols = ".#{name}"}
 
+  unless $install_procs.empty?
+    w = (w = ENV["COLUMNS"] and (w = w.to_i) > 80) ? w - 30 : 50
+    opt.on("\n""Types for --install and --exclude:")
+    mesg = +" "
+    $install_procs.each_key do |t|
+      if mesg.size + t.size > w
+        opt.on(mesg)
+        mesg = +" "
+      end
+      mesg << " " << t.to_s
+    end
+    opt.on(mesg)
+  end
+
   opt.order!(argv) do |v|
     case v
     when /\AINSTALL[-_]([-\w]+)=(.*)/
@@ -135,6 +150,7 @@ def parse_args(argv = ARGV)
   end
 
   $destdir ||= $mflags.defined?("DESTDIR")
+  $destdir = File.expand_path($destdir) unless $destdir.empty?
   if $extout ||= $mflags.defined?("EXTOUT")
     RbConfig.expand($extout)
   end
@@ -157,6 +173,16 @@ def parse_args(argv = ARGV)
   end
 end
 
+Compressors = {".gz"=>"gzip", ".bz2"=>"bzip2"}
+def Compressors.for(type)
+  ext = File.extname(type)
+  if compress = fetch(ext, nil)
+    [type.chomp(ext), ext, compress]
+  else
+    [type, *find {|_, z| system(z, in: IO::NULL, out: IO::NULL)}]
+  end
+end
+
 $install_procs = Hash.new {[]}
 def install?(*types, &block)
   unless types.delete(:nodefault)
@@ -205,15 +231,20 @@ def ln_sf(src, dest)
 end
 
 $made_dirs = {}
+
+def dir_creating(dir)
+  $made_dirs.fetch(dir) do
+    $made_dirs[dir] = true
+    $installed_list.puts(File.join(dir, "")) if $installed_list
+    yield if defined?(yield)
+  end
+end
+
 def makedirs(dirs)
   dirs = fu_list(dirs)
   dirs.collect! do |dir|
     realdir = with_destdir(dir)
-    realdir unless $made_dirs.fetch(dir) do
-      $made_dirs[dir] = true
-      $installed_list.puts(File.join(dir, "")) if $installed_list
-      File.directory?(realdir)
-    end
+    realdir unless dir_creating(dir) {File.directory?(realdir)}
   end.compact!
   super(dirs, :mode => $dir_mode) unless dirs.empty?
 end
@@ -346,6 +377,9 @@ rubyw_install_name = CONFIG["rubyw_install_name"]
 goruby_install_name = "go" + ruby_install_name
 
 bindir = CONFIG["bindir", true]
+if CONFIG["libdirname"] == "archlibdir"
+  archbindir = bindir.sub(%r[/\K(?=[^/]+\z)]) {CONFIG["config_target"] + "/"}
+end
 libdir = CONFIG[CONFIG.fetch("libdirname", "libdir"), true]
 rubyhdrdir = CONFIG["rubyhdrdir", true]
 archhdrdir = CONFIG["rubyarchhdrdir"] || (rubyhdrdir + "/" + CONFIG['arch'])
@@ -369,113 +403,13 @@ load_relative = CONFIG["LIBRUBY_RELATIVE"] == 'yes'
 
 rdoc_noinst = %w[created.rid]
 
-install?(:local, :arch, :bin, :'bin-arch') do
-  prepare "binary commands", bindir
-
-  install ruby_install_name+exeext, bindir, :mode => $prog_mode, :strip => $strip
-  if rubyw_install_name and !rubyw_install_name.empty?
-    install rubyw_install_name+exeext, bindir, :mode => $prog_mode, :strip => $strip
-  end
-  # emcc produces ruby and ruby.wasm, the first is a JavaScript file of runtime support
-  # to load and execute the second .wasm file. Both are required to execute ruby
-  if RUBY_PLATFORM =~ /emscripten/ and File.exist? ruby_install_name+".wasm"
-    install ruby_install_name+".wasm", bindir, :mode => $prog_mode, :strip => $strip
-  end
-  if File.exist? goruby_install_name+exeext
-    install goruby_install_name+exeext, bindir, :mode => $prog_mode, :strip => $strip
-  end
-  if enable_shared and dll != lib
-    install dll, bindir, :mode => $prog_mode, :strip => $strip
-  end
-end
-
-install?(:local, :arch, :lib, :'lib-arch') do
-  prepare "base libraries", libdir
-
-  install lib, libdir, :mode => $prog_mode, :strip => $strip unless lib == arc
-  install arc, libdir, :mode => $data_mode unless CONFIG["INSTALL_STATIC_LIBRARY"] == "no"
-  if dll == lib and dll != arc
-    for link in CONFIG["LIBRUBY_ALIASES"].split - [File.basename(dll)]
-      ln_sf(dll, File.join(libdir, link))
-    end
-  end
-
-  prepare "arch files", archlibdir
-  install "rbconfig.rb", archlibdir, :mode => $data_mode
-  if CONFIG["ARCHFILE"]
-    for file in CONFIG["ARCHFILE"].split
-      install file, archlibdir, :mode => $data_mode
-    end
-  end
-end
-
-install?(:local, :arch, :data) do
-  pc = CONFIG["ruby_pc"]
-  if pc and File.file?(pc) and File.size?(pc)
-    prepare "pkgconfig data", pkgconfigdir = File.join(libdir, "pkgconfig")
-    install pc, pkgconfigdir, :mode => $data_mode
-  end
-end
-
-install?(:ext, :arch, :'ext-arch') do
-  prepare "extension objects", archlibdir
-  noinst = %w[-* -*/] | (CONFIG["no_install_files"] || "").split
-  install_recursive("#{$extout}/#{CONFIG['arch']}", archlibdir, :no_install => noinst, :mode => $prog_mode, :strip => $strip)
-  prepare "extension objects", sitearchlibdir
-  prepare "extension objects", vendorarchlibdir
-  if extso = File.read("exts.mk")[/^EXTSO[ \t]*=[ \t]*((?:.*\\\n)*.*)/, 1] and
-    !(extso = extso.gsub(/\\\n/, '').split).empty?
-    libpathenv = CONFIG["LIBPATHENV"]
-    dest = CONFIG[!libpathenv || libpathenv == "PATH" ? "bindir" : "libdir"]
-    prepare "external libraries", dest
-    for file in extso
-      install file, dest, :mode => $prog_mode
-    end
-  end
-end
-install?(:ext, :arch, :hdr, :'arch-hdr', :'hdr-arch') do
-  prepare "extension headers", archhdrdir
-  install_recursive("#{$extout}/include/#{CONFIG['arch']}", archhdrdir, :glob => "*.h", :mode => $data_mode)
-  install_recursive("#{$extout}/include/#{CONFIG['arch']}", archhdrdir, :glob => "rb_rjit_header-*.obj", :mode => $data_mode)
-  install_recursive("#{$extout}/include/#{CONFIG['arch']}", archhdrdir, :glob => "rb_rjit_header-*.pch", :mode => $data_mode)
-end
-install?(:ext, :comm, :'ext-comm') do
-  prepare "extension scripts", rubylibdir
-  install_recursive("#{$extout}/common", rubylibdir, :mode => $data_mode)
-  prepare "extension scripts", sitelibdir
-  prepare "extension scripts", vendorlibdir
-end
-install?(:ext, :comm, :hdr, :'comm-hdr', :'hdr-comm') do
-  hdrdir = rubyhdrdir + "/ruby"
-  prepare "extension headers", hdrdir
-  install_recursive("#{$extout}/include/ruby", hdrdir, :glob => "*.h", :mode => $data_mode)
-end
-
-install?(:doc, :rdoc) do
-  if $rdocdir
-    ridatadir = File.join(CONFIG['ridir'], CONFIG['ruby_version'], "system")
-    prepare "rdoc", ridatadir
-    install_recursive($rdocdir, ridatadir, :no_install => rdoc_noinst, :mode => $data_mode)
-  end
-end
-install?(:doc, :html) do
-  if $htmldir
-    prepare "html-docs", docdir
-    install_recursive($htmldir, docdir+"/html", :no_install => rdoc_noinst, :mode => $data_mode)
-  end
-end
-install?(:doc, :capi) do
-  prepare "capi-docs", docdir
-  install_recursive "doc/capi", docdir+"/capi", :mode => $data_mode
-end
-
 prolog_script = <<EOS
 bindir="#{load_relative ? '${0%/*}' : bindir.gsub(/\"/, '\\\\"')}"
 EOS
-if CONFIG["LIBRUBY_RELATIVE"] != 'yes' and libpathenv = CONFIG["LIBPATHENV"]
+if !load_relative and libpathenv = CONFIG["LIBPATHENV"]
   pathsep = File::PATH_SEPARATOR
   prolog_script << <<EOS
-libdir="#{load_relative ? '$\{bindir%/bin\}/lib' : libdir.gsub(/\"/, '\\\\"')}"
+libdir="#{libdir.gsub(/\"/, '\\\\"')}"
 export #{libpathenv}="$libdir${#{libpathenv}:+#{pathsep}$#{libpathenv}}"
 EOS
 end
@@ -584,129 +518,6 @@ $script_installer = Class.new(installer) do
   break new(ruby_shebang, ruby_bin, ruby_install_name, nil, trans)
 end
 
-install?(:local, :comm, :bin, :'bin-comm') do
-  prepare "command scripts", bindir
-
-  install_recursive(File.join(srcdir, "bin"), bindir, :maxdepth => 1) do |src, cmd|
-    $script_installer.install(src, cmd)
-  end
-end
-
-install?(:local, :comm, :lib) do
-  prepare "library scripts", rubylibdir
-  noinst = %w[*.txt *.rdoc *.gemspec]
-  install_recursive(File.join(srcdir, "lib"), rubylibdir, :no_install => noinst, :mode => $data_mode)
-end
-
-install?(:local, :comm, :hdr, :'comm-hdr') do
-  prepare "common headers", rubyhdrdir
-
-  noinst = []
-  unless RUBY_PLATFORM =~ /mswin|mingw|bccwin/
-    noinst << "win32.h"
-  end
-  noinst = nil if noinst.empty?
-  install_recursive(File.join(srcdir, "include"), rubyhdrdir, :no_install => noinst, :glob => "*.{h,hpp}", :mode => $data_mode)
-end
-
-install?(:local, :comm, :man) do
-  mdocs = Dir["#{srcdir}/man/*.[1-9]"]
-  prepare "manpages", mandir, ([] | mdocs.collect {|mdoc| mdoc[/\d+$/]}).sort.collect {|sec| "man#{sec}"}
-
-  case $mantype
-  when /\.(?:(gz)|bz2)\z/
-    compress = $1 ? "gzip" : "bzip2"
-    suffix = $&
-  end
-  mandir = File.join(mandir, "man")
-  has_goruby = File.exist?(goruby_install_name+exeext)
-  require File.join(srcdir, "tool/mdoc2man.rb") if /\Adoc\b/ !~ $mantype
-  mdocs.each do |mdoc|
-    next unless File.file?(mdoc) and File.read(mdoc, 1) == '.'
-    base = File.basename(mdoc)
-    if base == "goruby.1"
-      next unless has_goruby
-    end
-
-    destdir = mandir + (section = mdoc[/\d+$/])
-    destname = ruby_install_name.sub(/ruby/, base.chomp(".#{section}"))
-    destfile = File.join(destdir, "#{destname}.#{section}")
-
-    if /\Adoc\b/ =~ $mantype
-      if compress
-        begin
-          w = IO.popen(compress, "rb", in: mdoc, &:read)
-        rescue
-        else
-          destfile << suffix
-        end
-      end
-      if w
-        open_for_install(destfile, $data_mode) {w}
-      else
-        install mdoc, destfile, :mode => $data_mode
-      end
-    else
-      class << (w = [])
-        alias print push
-      end
-      if File.basename(mdoc).start_with?('bundle') ||
-         File.basename(mdoc).start_with?('gemfile')
-        w = File.read(mdoc)
-      else
-        File.open(mdoc) {|r| Mdoc2Man.mdoc2man(r, w)}
-        w = w.join("")
-      end
-      if compress
-        begin
-          w = IO.popen(compress, "r+b") do |f|
-            Thread.start {f.write w; f.close_write}
-            f.read
-          end
-        rescue
-        else
-          destfile << suffix
-        end
-      end
-      open_for_install(destfile, $data_mode) {w}
-    end
-  end
-end
-
-install?(:dbg, :nodefault) do
-  prepare "debugger commands", bindir
-  prepare "debugger scripts", rubylibdir
-  conf = RbConfig::MAKEFILE_CONFIG.merge({"prefix"=>"${prefix#/}"})
-  Dir.glob(File.join(srcdir, "template/ruby-*db.in")) do |src|
-    cmd = $script_installer.transform(File.basename(src, ".in"))
-    open_for_install(File.join(bindir, cmd), $script_mode) {
-      RbConfig.expand(File.read(src), conf)
-    }
-  end
-  Dir.glob(File.join(srcdir, "misc/lldb_*")) do |src|
-    if File.directory?(src)
-      install_recursive src, File.join(rubylibdir, File.basename(src))
-    else
-      install src, rubylibdir
-    end
-  end
-  install File.join(srcdir, ".gdbinit"), File.join(rubylibdir, "gdbinit")
-  if $debug_symbols
-    {
-      ruby_install_name => bindir,
-      rubyw_install_name => bindir,
-      goruby_install_name => bindir,
-      dll => libdir,
-    }.each do |src, dest|
-      next if src.empty?
-      src += $debug_symbols
-      if File.directory?(src)
-        install_recursive src, File.join(dest, src)
-      end
-    end
-  end
-end
-
 module RbInstall
   def self.no_write(options = nil)
     u = File.umask(0022)
@@ -716,6 +527,8 @@ module RbInstall
         const_set(:FileUtils, fu::NoWrite)
         fu
       end
+      # RubyGems 3.0.0 or later supports `dir_mode`, but it uses
+      # `File` method to apply it, not `FileUtils`.
       dir_mode = options.delete(:dir_mode) if options
     end
     yield
@@ -745,47 +558,119 @@ module RbInstall
       end
 
       def collect
-        ruby_libraries.sort
+        requirable_features.sort
+      end
+
+      private
+
+      def features_from_makefile(makefile_path)
+        makefile = File.read(makefile_path)
+
+        name = makefile[/^TARGET[ \t]*=[ \t]*((?:.*\\\n)*.*)/, 1]
+        return [] if name.nil? || name.empty?
+
+        feature = makefile[/^DLLIB[ \t]*=[ \t]*((?:.*\\\n)*.*)/, 1]
+        feature = feature.sub("$(TARGET)", name)
+
+        target_prefix = makefile[/^target_prefix[ \t]*=[ \t]*((?:.*\\\n)*.*)/, 1]
+        feature = File.join(target_prefix.delete_prefix("/"), feature) unless target_prefix.empty?
+
+        Array(feature)
       end
 
       class Ext < self
-        def skip_install?(files)
+        def requirable_features
           # install ext only when it's configured
-          !File.exist?("#{$ext_build_dir}/#{relative_base}/Makefile")
+          return [] unless File.exist?(makefile_path)
+
+          ruby_features + ext_features
         end
 
-        def ruby_libraries
-          Dir.glob("lib/**/*.rb", base: "#{srcdir}/ext/#{relative_base}")
+        private
+
+        def ruby_features
+          Dir.glob("**/*.rb", base: "#{makefile_dir}/lib")
+        end
+
+        def ext_features
+          features_from_makefile(makefile_path)
+        end
+
+        def makefile_path
+          if File.exist?("#{makefile_dir}/Makefile")
+            "#{makefile_dir}/Makefile"
+          else
+            # for out-of-place build
+            "#{$ext_build_dir}/#{relative_base}/Makefile"
+          end
+        end
+
+        def makefile_dir
+          "#{root}/#{relative_base}"
+        end
+
+        def root
+          File.expand_path($ext_build_dir, srcdir)
         end
       end
 
       class Lib < self
-        def skip_install?(files)
-          files.empty?
+        def requirable_features
+          ruby_features + ext_features
         end
 
-        def ruby_libraries
+        private
+
+        def ruby_features
           gemname = File.basename(gemspec, ".gemspec")
           base = relative_base || gemname
           # for lib/net/net-smtp.gemspec
           if m = /.*(?=-(.*)\z)/.match(gemname)
             base = File.join(base, *m.to_a.select {|n| !base.include?(n)})
           end
-          files = Dir.glob("lib/#{base}{.rb,/**/*.rb}", base: srcdir)
+          files = Dir.glob("#{base}{.rb,/**/*.rb}", base: root)
           if !relative_base and files.empty? # no files at the toplevel
             # pseudo gem like ruby2_keywords
-            files << "lib/#{gemname}.rb"
+            files << "#{gemname}.rb"
           end
 
           case gemname
           when "net-http"
-            files << "lib/net/https.rb"
+            files << "net/https.rb"
           when "optparse"
-            files << "lib/optionparser.rb"
+            files << "optionparser.rb"
           end
 
           files
         end
+
+        def ext_features
+          loaded_gemspec = load_gemspec("#{root}/#{gemspec}")
+          extension = loaded_gemspec.extensions.first
+          return [] unless extension
+
+          extconf = File.expand_path(extension, srcdir)
+          ext_build_dir = File.dirname(extconf)
+          makefile_path = "#{ext_build_dir}/Makefile"
+          return [] unless File.exist?(makefile_path)
+
+          features_from_makefile(makefile_path)
+        end
+
+        def root
+          "#{srcdir}/lib"
+        end
+      end
+
+      class UnpackedGem < self
+        def collect
+          base = @srcdir or return []
+          Dir.glob("**/*", File::FNM_DOTMATCH, base: base).select do |n|
+            next if n == "."
+            next if File.fnmatch?("*.gemspec", n, File::FNM_DOTMATCH|File::FNM_PATHNAME)
+            !File.directory?(File.join(base, n))
+          end
+        end
       end
     end
   end
@@ -823,10 +708,78 @@ module RbInstall
     end
   end
 
-  class GemInstaller < Gem::Installer
-  end
+  class UnpackedInstaller < Gem::Installer
+    # This method is mostly copied from old version of Gem::Installer#install
+    def install_with_default_gem
+      verify_gem_home
+
+      # The name and require_paths must be verified first, since it could contain
+      # ruby code that would be eval'ed in #ensure_loadable_spec
+      verify_spec
+
+      ensure_loadable_spec
+
+      if options[:install_as_default]
+        Gem.ensure_default_gem_subdirectories gem_home
+      else
+        Gem.ensure_gem_subdirectories gem_home
+      end
+
+      return true if @force
+
+      ensure_dependencies_met unless @ignore_dependencies
+
+      run_pre_install_hooks
+
+      # Set loaded_from to ensure extension_dir is correct
+      if @options[:install_as_default]
+        spec.loaded_from = default_spec_file
+      else
+        spec.loaded_from = spec_file
+      end
+
+      # Completely remove any previous gem files
+      FileUtils.rm_rf gem_dir
+      FileUtils.rm_rf spec.extension_dir
+
+      dir_mode = options[:dir_mode]
+      FileUtils.mkdir_p gem_dir, mode: dir_mode && 0o755
+
+      if @options[:install_as_default]
+        extract_bin
+        write_default_spec
+      else
+        extract_files
+
+        build_extensions
+        write_build_info_file
+        run_post_build_hooks
+      end
+
+      generate_bin
+      generate_plugins
+
+      unless @options[:install_as_default]
+        write_spec
+        write_cache_file
+      end
+
+      File.chmod(dir_mode, gem_dir) if dir_mode
+
+      say spec.post_install_message if options[:post_install_message] && !spec.post_install_message.nil?
+
+      Gem::Specification.add_spec(spec) unless @install_dir
+
+      load_plugin
+
+      run_post_install_hooks
+
+      spec
+    rescue Errno::EACCES => e
+      # Permission denied - /path/to/foo
+      raise Gem::FilePermissionError, e.message.split(" - ").last
+    end
 
-  class UnpackedInstaller < GemInstaller
     def write_cache_file
     end
 
@@ -850,11 +803,6 @@ module RbInstall
       super
     end
 
-    def generate_bin_script(filename, bindir)
-      return if same_bin_script?(filename, bindir)
-      super
-    end
-
     def same_bin_script?(filename, bindir)
       path = File.join(bindir, formatted_program_filename(filename))
       begin
@@ -873,12 +821,11 @@ module RbInstall
       super unless $dryrun
       $installed_list.puts(without_destdir(default_spec_file)) if $installed_list
     end
-  end
 
-  class GemInstaller
     def install
       spec.post_install_message = nil
-      RbInstall.no_write(options) {super}
+      dir_creating(without_destdir(gem_dir))
+      RbInstall.no_write(options) { install_with_default_gem }
     end
 
     # Now build-ext builds all extensions including bundled gems.
@@ -886,10 +833,13 @@ module RbInstall
     end
 
     def generate_bin_script(filename, bindir)
+      return if same_bin_script?(filename, bindir)
       name = formatted_program_filename(filename)
       unless $dryrun
         super
-        File.chmod($script_mode, File.join(bindir, name))
+        script = File.join(bindir, name)
+        File.chmod($script_mode, script)
+        File.unlink("#{script}.lock") rescue nil
       end
       $installed_list.puts(File.join(without_destdir(bindir), name)) if $installed_list
     end
@@ -904,38 +854,36 @@ module RbInstall
         $installed_list.puts(d+"/") if $installed_list
       end
     end
+
+    def load_plugin
+      # Suppress warnings for constant re-assignment
+      verbose, $VERBOSE = $VERBOSE, nil
+      super
+    ensure
+      $VERBOSE = verbose
+    end
   end
 end
 
-# :startdoc:
-
-install?(:ext, :comm, :gem, :'default-gems', :'default-gems-comm') do
-  install_default_gem('lib', srcdir, bindir)
-end
-install?(:ext, :arch, :gem, :'default-gems', :'default-gems-arch') do
-  install_default_gem('ext', srcdir, bindir)
-end
-
-def load_gemspec(file, base = nil)
+def load_gemspec(file, base = nil, files: nil)
   file = File.realpath(file)
   code = File.read(file, encoding: "utf-8:-")
-  code.gsub!(/(?:`git[^\`]*`|%x\[git[^\]]*\])\.split\([^\)]*\)/m) do
-    files = []
-    if base
-      Dir.glob("**/*", File::FNM_DOTMATCH, base: base) do |n|
-        case File.basename(n); when ".", ".."; next; end
-        next if File.directory?(File.join(base, n))
-        files << n.dump
-      end
-    end
-    "[" + files.join(", ") + "]"
+
+  code.gsub!(/^ *#.*/, "")
+  files = files ? files.map(&:dump).join(", ") : ""
+  code.gsub!(/(?:`git[^\`]*`|%x\[git[^\]]*\])\.split(\([^\)]*\))?/m) do
+    "[" + files + "]"
+  end \
+  or
+  code.gsub!(/IO\.popen\(.*git.*?\)/) do
+    "[" + files + "] || itself"
   end
+
   spec = eval(code, binding, file)
   unless Gem::Specification === spec
     raise TypeError, "[#{file}] isn't a Gem::Specification (#{spec.class} instead)."
   end
   spec.loaded_from = base ? File.join(base, File.basename(file)) : file
-  spec.files.reject! {|n| n.end_with?(".gemspec") or n.start_with?(".git")}
   spec.date = RUBY_RELEASE_DATE
 
   spec
@@ -946,6 +894,7 @@ def install_default_gem(dir, srcdir, bindir)
   install_dir = with_destdir(gem_dir)
   prepare "default gems from #{dir}", gem_dir
   RbInstall.no_write do
+    # Record making directories
     makedirs(Gem.ensure_default_gem_subdirectories(install_dir, $dir_mode).map {|d| File.join(gem_dir, d)})
   end
 
@@ -964,14 +913,11 @@ def install_default_gem(dir, srcdir, bindir)
 
   base = "#{srcdir}/#{dir}"
   gems = Dir.glob("**/*.gemspec", base: base).map {|src|
-    spec = load_gemspec("#{base}/#{src}")
-    file_collector = RbInstall::Specs::FileCollector.for(srcdir, dir, src)
-    files = file_collector.collect
-    if file_collector.skip_install?(files)
+    files = RbInstall::Specs::FileCollector.for(srcdir, dir, src).collect
+    if files.empty?
       next
     end
-    spec.files = files
-    spec
+    load_gemspec("#{base}/#{src}", files: files)
   }
   gems.compact.sort_by(&:name).each do |gemspec|
     old_gemspecs = Dir[File.join(with_destdir(default_spec_dir), "#{gemspec.name}-*.gemspec")]
@@ -990,11 +936,259 @@ def install_default_gem(dir, srcdir, bindir)
   end
 end
 
+def mdoc_file?(mdoc)
+  /^\.Nm / =~ File.read(mdoc, 1024)
+end
+
+# :startdoc:
+
+install?(:local, :arch, :bin, :'bin-arch') do
+  prepare "binary commands", (dest = archbindir || bindir)
+
+  def (bins = []).add(name)
+    push(name)
+    name
+  end
+
+  install bins.add(ruby_install_name+exeext), dest, :mode => $prog_mode, :strip => $strip
+  if rubyw_install_name and !rubyw_install_name.empty?
+    install bins.add(rubyw_install_name+exeext), dest, :mode => $prog_mode, :strip => $strip
+  end
+  # emcc produces ruby and ruby.wasm, the first is a JavaScript file of runtime support
+  # to load and execute the second .wasm file. Both are required to execute ruby
+  if RUBY_PLATFORM =~ /emscripten/ and File.exist? ruby_install_name+".wasm"
+    install bins.add(ruby_install_name+".wasm"), dest, :mode => $prog_mode, :strip => $strip
+  end
+  if File.exist? goruby_install_name+exeext
+    install bins.add(goruby_install_name+exeext), dest, :mode => $prog_mode, :strip => $strip
+  end
+  if enable_shared and dll != lib
+    install bins.add(dll), dest, :mode => $prog_mode, :strip => $strip
+  end
+  if archbindir
+    prepare "binary command links", bindir
+    relpath = Path.relative(archbindir, bindir)
+    bins.each do |f|
+      ln_sf(File.join(relpath, f), File.join(bindir, f))
+    end
+  end
+end
+
+install?(:local, :arch, :lib, :'lib-arch') do
+  prepare "base libraries", libdir
+
+  install lib, libdir, :mode => $prog_mode, :strip => $strip unless lib == arc
+  install arc, libdir, :mode => $data_mode unless CONFIG["INSTALL_STATIC_LIBRARY"] == "no"
+  if dll == lib and dll != arc
+    for link in CONFIG["LIBRUBY_ALIASES"].split - [File.basename(dll)]
+      ln_sf(dll, File.join(libdir, link))
+    end
+  end
+
+  prepare "arch files", archlibdir
+  install "rbconfig.rb", archlibdir, :mode => $data_mode
+  if CONFIG["ARCHFILE"]
+    for file in CONFIG["ARCHFILE"].split
+      install file, archlibdir, :mode => $data_mode
+    end
+  end
+end
+
+install?(:local, :arch, :data) do
+  pc = CONFIG["ruby_pc"]
+  if pc and File.file?(pc) and File.size?(pc)
+    prepare "pkgconfig data", pkgconfigdir = File.join(libdir, "pkgconfig")
+    install pc, pkgconfigdir, :mode => $data_mode
+    if (pkgconfig_base = CONFIG["libdir", true]) != libdir
+      prepare "pkgconfig data link", File.join(pkgconfig_base, "pkgconfig")
+      ln_sf(File.join("..", Path.relative(pkgconfigdir, pkgconfig_base), pc),
+            File.join(pkgconfig_base, "pkgconfig", pc))
+    end
+  end
+end
+
+install?(:ext, :arch, :'ext-arch') do
+  prepare "extension objects", archlibdir
+  noinst = %w[-* -*/] | (CONFIG["no_install_files"] || "").split
+  install_recursive("#{$extout}/#{CONFIG['arch']}", archlibdir, :no_install => noinst, :mode => $prog_mode, :strip => $strip)
+  prepare "extension objects", sitearchlibdir
+  prepare "extension objects", vendorarchlibdir
+  if extso = File.read("exts.mk")[/^EXTSO[ \t]*=[ \t]*((?:.*\\\n)*.*)/, 1] and
+    !(extso = extso.gsub(/\\\n/, '').split).empty?
+    libpathenv = CONFIG["LIBPATHENV"]
+    dest = CONFIG[!libpathenv || libpathenv == "PATH" ? "bindir" : "libdir"]
+    prepare "external libraries", dest
+    for file in extso
+      install file, dest, :mode => $prog_mode
+    end
+  end
+end
+
+install?(:ext, :arch, :hdr, :'arch-hdr', :'hdr-arch') do
+  prepare "extension headers", archhdrdir
+  install_recursive("#{$extout}/include/#{CONFIG['arch']}", archhdrdir, :glob => "*.h", :mode => $data_mode)
+end
+
+install?(:ext, :comm, :'ext-comm') do
+  prepare "extension scripts", rubylibdir
+  install_recursive("#{$extout}/common", rubylibdir, :mode => $data_mode)
+  prepare "extension scripts", sitelibdir
+  prepare "extension scripts", vendorlibdir
+end
+
+install?(:ext, :comm, :hdr, :'comm-hdr', :'hdr-comm') do
+  hdrdir = rubyhdrdir + "/ruby"
+  prepare "extension headers", hdrdir
+  install_recursive("#{$extout}/include/ruby", hdrdir, :glob => "*.h", :mode => $data_mode)
+end
+
+install?(:doc, :rdoc) do
+  if $rdocdir
+    ridatadir = File.join(CONFIG['ridir'], CONFIG['ruby_version'], "system")
+    prepare "rdoc", ridatadir
+    install_recursive($rdocdir, ridatadir, :no_install => rdoc_noinst, :mode => $data_mode)
+  end
+end
+
+install?(:doc, :html) do
+  if $htmldir
+    prepare "html-docs", docdir
+    install_recursive($htmldir, docdir+"/html", :no_install => rdoc_noinst, :mode => $data_mode)
+  end
+end
+
+install?(:doc, :capi) do
+  prepare "capi-docs", docdir
+  install_recursive "doc/capi", docdir+"/capi", :mode => $data_mode
+end
+
+install?(:local, :comm, :bin, :'bin-comm') do
+  prepare "command scripts", bindir
+
+  install_recursive(File.join(srcdir, "bin"), bindir, :maxdepth => 1) do |src, cmd|
+    $script_installer.install(src, cmd)
+  end
+end
+
+install?(:local, :comm, :lib) do
+  prepare "library scripts", rubylibdir
+  noinst = %w[*.txt *.rdoc *.gemspec]
+  install_recursive(File.join(srcdir, "lib"), rubylibdir, :no_install => noinst, :mode => $data_mode)
+end
+
+install?(:local, :comm, :hdr, :'comm-hdr') do
+  prepare "common headers", rubyhdrdir
+
+  noinst = []
+  unless RUBY_PLATFORM =~ /mswin|mingw|bccwin/
+    noinst << "win32.h"
+  end
+  noinst = nil if noinst.empty?
+  install_recursive(File.join(srcdir, "include"), rubyhdrdir, :no_install => noinst, :glob => "*.{h,hpp}", :mode => $data_mode)
+end
+
+install?(:local, :comm, :man) do
+  mdocs = Dir["#{srcdir}/man/*.[1-9]"]
+  prepare "manpages", mandir, ([] | mdocs.collect {|mdoc| mdoc[/\d+$/]}).sort.collect {|sec| "man#{sec}"}
+
+  mantype, suffix, compress = Compressors.for($mantype)
+  has_goruby = File.exist?(goruby_install_name+exeext)
+  require File.join(srcdir, "tool/mdoc2man.rb") if /\Adoc\b/ !~ mantype
+  mdocs.each do |mdoc|
+    next unless File.file?(mdoc) and File.read(mdoc, 1) == '.'
+    base = File.basename(mdoc)
+    if base == "goruby.1"
+      next unless has_goruby
+    end
+
+    destdir = File.join(mandir, "man" + (section = mdoc[/\d+$/]))
+    destname = $script_installer.transform(base.chomp(".#{section}"))
+    destfile = File.join(destdir, "#{destname}.#{section}")
+
+    if /\Adoc\b/ =~ mantype or !mdoc_file?(mdoc)
+      if compress
+        begin
+          w = IO.popen(compress, "rb", in: mdoc, &:read)
+        rescue
+        else
+          destfile << suffix
+        end
+      end
+      if w
+        open_for_install(destfile, $data_mode) {w}
+      else
+        install mdoc, destfile, :mode => $data_mode
+      end
+    else
+      class << (w = [])
+        alias print push
+      end
+      File.open(mdoc) {|r| Mdoc2Man.mdoc2man(r, w)}
+      w = w.join("")
+      if compress
+        begin
+          w = IO.popen(compress, "r+b") do |f|
+            Thread.start {f.write w; f.close_write}
+            f.read
+          end
+        rescue
+        else
+          destfile << suffix
+        end
+      end
+      open_for_install(destfile, $data_mode) {w}
+    end
+  end
+end
+
+install?(:dbg, :nodefault) do
+  prepare "debugger commands", bindir
+  prepare "debugger scripts", rubylibdir
+  conf = MAKEFILE_CONFIG.merge({"prefix"=>"${prefix#/}"})
+  Dir.glob(File.join(srcdir, "template/ruby-*db.in")) do |src|
+    cmd = $script_installer.transform(File.basename(src, ".in"))
+    open_for_install(File.join(bindir, cmd), $script_mode) {
+      RbConfig.expand(File.read(src), conf)
+    }
+  end
+  Dir.glob(File.join(srcdir, "misc/lldb_*")) do |src|
+    if File.directory?(src)
+      install_recursive src, File.join(rubylibdir, File.basename(src))
+    else
+      install src, rubylibdir
+    end
+  end
+  install File.join(srcdir, ".gdbinit"), File.join(rubylibdir, "gdbinit")
+  if $debug_symbols
+    {
+      ruby_install_name => archbindir || bindir,
+      rubyw_install_name => archbindir || bindir,
+      goruby_install_name => archbindir || bindir,
+      dll => libdir,
+    }.each do |src, dest|
+      next if src.empty?
+      src += $debug_symbols
+      if File.directory?(src)
+        install_recursive src, File.join(dest, src)
+      end
+    end
+  end
+end
+
+install?(:ext, :comm, :gem, :'default-gems', :'default-gems-comm') do
+  install_default_gem('lib', srcdir, bindir)
+end
+
+install?(:ext, :arch, :gem, :'default-gems', :'default-gems-arch') do
+  install_default_gem('ext', srcdir, bindir)
+end
+
 install?(:ext, :comm, :gem, :'bundled-gems') do
   gem_dir = Gem.default_dir
   install_dir = with_destdir(gem_dir)
   prepare "bundled gems", gem_dir
   RbInstall.no_write do
+    # Record making directories
     makedirs(Gem.ensure_gem_subdirectories(install_dir, $dir_mode).map {|d| File.join(gem_dir, d)})
   end
 
@@ -1023,22 +1217,30 @@ install?(:ext, :comm, :gem, :'bundled-gems') do
   # the newly installed ruby.
   ENV.delete('RUBYOPT')
 
+  collector = RbInstall::Specs::FileCollector::UnpackedGem
   File.foreach("#{srcdir}/gems/bundled_gems") do |name|
     next if /^\s*(?:#|$)/ =~ name
     next unless /^(\S+)\s+(\S+).*/ =~ name
+    gem = $1
     gem_name = "#$1-#$2"
-    # Try to find the gemspec file for C ext gems
-    # ex .bundle/gems/debug-1.7.1/debug-1.7.1.gemspec
-    # This gemspec keep the original dependencies
-    path = "#{srcdir}/.bundle/gems/#{gem_name}/#{gem_name}.gemspec"
-    unless File.exist?(path)
-      path = "#{srcdir}/.bundle/specifications/#{gem_name}.gemspec"
-      unless File.exist?(path)
-         skipped[gem_name] = "gemspec not found"
-         next
-      end
+    path = [
+      # gemspec that removed duplicated dependencies of bundled gems
+      "#{srcdir}/.bundle/gems/#{gem_name}/#{gem}.gemspec",
+      # gemspec for C ext gems, It has the original dependencies
+      # ex .bundle/gems/debug-1.7.1/debug-1.7.1.gemspec
+      "#{srcdir}/.bundle/gems/#{gem_name}/#{gem_name}.gemspec",
+      # original gemspec generated by rubygems
+      "#{srcdir}/.bundle/specifications/#{gem_name}.gemspec"
+    ].find { |gemspec| File.exist?(gemspec) }
+    if path.nil?
+      skipped[gem_name] = "gemspec not found"
+      next
     end
-    spec = load_gemspec(path, "#{srcdir}/.bundle/gems/#{gem_name}")
+    base = "#{srcdir}/.bundle/gems/#{gem_name}"
+    files = collector.new(path, base, nil).collect
+    files.delete("#{gem}.gemspec")
+    files.delete("#{gem_name}.gemspec")
+    spec = load_gemspec(path, base, files: files)
     unless spec.platform == Gem::Platform::RUBY
       skipped[gem_name] = "not ruby platform (#{spec.platform})"
       next
@@ -1047,7 +1249,13 @@ install?(:ext, :comm, :gem, :'bundled-gems') do
       skipped[gem_name] = "full name unmatch #{spec.full_name}"
       next
     end
+    # Skip install C ext bundled gem if it is build failed or not found
+    if !spec.extensions.empty? && !File.exist?("#{build_dir}/#{gem_name}/gem.build_complete")
+      skipped[gem_name] = "extensions not found or build failed #{spec.full_name}"
+      next
+    end
     spec.extension_dir = "#{extensions_dir}/#{spec.full_name}"
+
     package = RbInstall::DirPackage.new spec
     ins = RbInstall::UnpackedInstaller.new(package, options)
     puts "#{INDENT}#{spec.name} #{spec.version}"
@@ -1067,11 +1275,21 @@ install?(:ext, :comm, :gem, :'bundled-gems') do
     skipped.default = "not found in bundled_gems"
     puts "skipped bundled gems:"
     gems.each do |gem|
-      printf "    %-32s%s\n", File.basename(gem), skipped[gem]
+      gem = File.basename(gem)
+      printf "    %-31s %s\n", gem, skipped[gem.chomp(".gem")]
     end
   end
 end
 
+install?('modular-gc') do
+  if modular_gc_dir = CONFIG['modular_gc_dir'] and !modular_gc_dir.empty?
+    dlext = CONFIG['DLEXT', true]
+    modular_gc_dir = File.expand_path(modular_gc_dir, CONFIG['prefix'])
+    prepare "modular GC library", modular_gc_dir
+    install Dir.glob("gc/*/librubygc.*.#{dlext}"), modular_gc_dir
+  end
+end
+
 parse_args()
 
 include FileUtils
@@ -1088,7 +1306,6 @@ installs = $install.map do |inst|
 end
 installs.flatten!
 installs -= $exclude.map {|exc| $install_procs[exc]}.flatten
-puts "Installing to #$destdir" unless installs.empty?
 installs.each do |block|
   dir = Dir.pwd
   begin
@@ -1097,5 +1314,9 @@ installs.each do |block|
     Dir.chdir(dir)
   end
 end
+unless installs.empty? or $destdir.empty?
+  require_relative 'lib/colorize'
+  puts "Installed under #{Colorize.new.info($destdir)}"
+end
 
 # vi:set sw=2:
